package ssw.mj.test;

import java.io.StringReader;

import junit.framework.TestCase;

import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.Token;
import ssw.mj.codegen.Code;
import ssw.mj.symtab.Obj;
import ssw.mj.symtab.Struct;
import ssw.mj.symtab.Tab;

/**
 * @author AW
 */
public class UEBeispieleTest extends TestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";
	
	private static final int
		none       = Token.none,
		ident      = Token.ident,
		number     = Token.number,
		charConst  = Token.charConst,
		plus       = Token.plus,
		minus      = Token.minus,
		times      = Token.times,
		slash      = Token.slash,
		rem        = Token.rem,
		eql        = Token.eql,
		neq        = Token.neq,
		gtr        = Token.gtr,
		geq        = Token.geq,
		less       = Token.less,
		leq        = Token.leq,
		and        = Token.and,
		or         = Token.or,
		assign     = Token.assign,
		pplus      = Token.pplus,
		mminus     = Token.mminus,
		semicolon  = Token.semicolon,
		comma      = Token.comma,
		period     = Token.period,
		lpar       = Token.lpar,
		rpar       = Token.rpar,
		lbrack     = Token.lbrack,
		rbrack     = Token.rbrack,
		lbrace     = Token.lbrace,
		rbrace     = Token.rbrace,
		classKW    = Token.classKW,
		continueKW = Token.continueKW,
		doKW       = Token.doKW,
		elseKW     = Token.elseKW,
		finalKW    = Token.finalKW,
		ifKW       = Token.ifKW,
		newKW      = Token.newKW,
		printKW    = Token.printKW,
		readKW     = Token.readKW,
		returnKW   = Token.returnKW,
		voidKW     = Token.voidKW,
		whileKW    = Token.whileKW,
		eof        = Token.eof;

	private static TestPrintWriter output;
	private static TestByteArrayOutputStream code;

	// ---------- helpers ----------
	
	private void initParser (String program) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(program), null);
		code = new TestByteArrayOutputStream();
	}

	public void testBspSymTab () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		////////////////////////		
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("A");
		expected = new Obj(Obj.Prog, "A", Tab.noType, 0, 0, null);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'A'", expected, actual);
		
		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages
		
		actual = actual.locals;
		expected = new Obj(Obj.Con, "max", Tab.intType, 12, 0, null);
		assertEquals("Constant 'max'", expected, actual);
				
		actual = actual.next;
		expected = new Obj(Obj.Var, "c", Tab.charType, 0, 0, null);
		assertEquals("Global 'c'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Var, "i", Tab.intType, 1, 0, null);
		assertEquals("Variable 'i'", expected, actual);

		actual = actual.next;
		Struct typeB = new Struct(Struct.Class);
		typeB.n = 2;
		typeB.fields = new Obj(Obj.Fld, "x", Tab.intType, 0, 1, null);
		assertEquals("Field 'x'", typeB.fields, actual.type.fields);
		typeB.fields.next = new Obj(Obj.Fld, "y", Tab.intType, 1, 1, null);
		assertEquals("Field 'y'", typeB.fields.next, actual.type.fields.next);
		expected = new Obj(Obj.Type, "B", typeB, 0, 0, null);
		assertEquals("Class 'B'", expected, actual);

		actual = actual.next;
		Obj locals = new Obj(Obj.Var, "iarr", new Struct(Struct.Arr, Tab.intType), 0, 1, null);
		assertEquals("Local 'iarr'", locals, actual.locals);
		locals.next = new Obj(Obj.Var, "b", typeB, 1, 1, null);
		assertEquals("Local 'b'", locals.next, actual.locals.next);
		locals.next.next = new Obj(Obj.Var, "n", Tab.intType, 2, 1, null);
		assertEquals("Local 'n'", locals.next.next, actual.locals.next.next);
		expected = new Obj(Obj.Meth, "main", Tab.noType, 0, 0, locals);
		assertEquals("Methode 'main'", expected, actual);
	}

	public void testBsp01 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    i = 10;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
			//   i = 10;
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(12, 1);  // putstatic
		code.expect( 1, 2);  // s
			// }

		Code.write(code);

		code.setCodeSize(8);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}
	
	public void testBsp02 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n = 3;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
			//   n = 3;
		code.expect(18, 1);  // const_3
		code.expect( 9, 1);  // store_2
			// }

		Code.write(code);

		code.setCodeSize(2);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}
	
	public void testBsp03 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n = 3 + i;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
			//   n = 3 + i;
		code.expect(18, 1);  // const_3
		code.expect(11, 1);  // getstatic
		code.expect( 1, 2);  // s
		code.expect(23, 1);  // add
		code.expect( 9, 1);  // store_2
			// }

		Code.write(code);

		code.setCodeSize(6);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}
	
	public void testBsp04 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n = 3 + i * max - n;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
			//   n = 3 + i * max - n;
		code.expect(18, 1);  // const_3
		code.expect(11, 1);  // getstatic
		code.expect( 1, 2);  // s
		code.expect(22, 1);  // const
		code.expect(12, 4);  // w
		code.expect(25, 1);  // mul
		code.expect(23, 1);  // add
		code.expect( 4, 1);  // load_2
		code.expect(24, 1);  // sub
		code.expect( 9, 1);  // store_2
			// }

		Code.write(code);

		code.setCodeSize(14);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}

	public void testBsp05 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    iarr = new int[10];" + LF +
				"    iarr[5] = 10;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
		  //   iarr = new int[10];
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(33, 1);  // newarray
		code.expect( 1, 1);  // b
		code.expect( 7, 1);  // store_0
			//   iarr[5] = 10;
		code.expect( 2, 1);  // load_0
		code.expect(20, 1);  // const_5
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(35, 1);  // astore
			// }

		Code.write(code);

		code.setCodeSize(16);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}
	
	public void testBsp06 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    iarr = new int[10];" + LF +
				"    iarr[5] = 10;" + LF +
				"    b = new B;" + LF +
				"    b.y = iarr[5] * 3;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
		  //   iarr = new int[10];
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(33, 1);  // newarray
		code.expect( 1, 1);  // b
		code.expect( 7, 1);  // store_0
			//   iarr[5] = 10;
		code.expect( 2, 1);  // load_0
		code.expect(20, 1);  // const_5
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(35, 1);  // astore
			//   b = new B;
		code.expect(32, 1);  // new
		code.expect( 8, 2);  // s
		code.expect( 8, 1);  // store_1
			//   b.y = iarr[5] * 3;
		code.expect( 3, 1);  // load_1
		code.expect( 2, 1);  // load_0
		code.expect(20, 1);  // const_5
		code.expect(34, 1);  // aload
		code.expect(18, 1);  // const_3
		code.expect(25, 1);  // mul
		code.expect(14, 1);  // putfield
		code.expect( 1, 2);  // s
			// }

		Code.write(code);

		code.setCodeSize(29);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}

	public void testBsp07 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    n--;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
		  //   n--;
		code.expect(31, 1);  // inc
		code.expect( 2, 1);  // b1
		code.expect(-1, 1);  // b2
			// }

		Code.write(code);

		code.setCodeSize(3);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}

	public void testBsp08 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    i++;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
		  //   i++;
		code.expect(11, 1);  // getstatic
		code.expect( 1, 2);  // s
		code.expect(16, 1);  // const_1
		code.expect(23, 1);  // add
		code.expect(12, 1);  // putstatic
		code.expect( 1, 2);  // s
			// }

		Code.write(code);

		code.setCodeSize(8);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}

	public void testBsp09 () {
		initParser(
				"class A" + LF +
				"  final int max = 12;" + LF +
				"  char c; int i;" + LF +
				"  class B { int x, y; }" + LF +
				"{" + LF +
				"  void main ()" + LF +
				"    int[] iarr; B b; int n;" + LF +
				"  {" + LF +
				"    iarr = new int[10];" + LF +
				"    iarr[0]--;" + LF +
				"  }" + LF +
				"}");
		
		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
			// void main () 
			//   int[] iarr; B b; int n;
			// {
		  //   iarr = new int[10];
		code.expect(22, 1);  // const
		code.expect(10, 4);  // w
		code.expect(33, 1);  // newarray
		code.expect( 1, 1);  // b
		code.expect( 7, 1);  // store_0
			//   iarr[0]--;
		code.expect( 2, 1);  // load_0
		code.expect(15, 1);  // const_0
		code.expect(41, 1);  // dup2
		code.expect(34, 1);  // aload
		code.expect(21, 1);  // const_m1
		code.expect(23, 1);  // add
		code.expect(35, 1);  // astore
			// }

		Code.write(code);

		code.setCodeSize(15);
		code.setDataSize(2);
		code.setMainPC(-1);
//		code.dump();		// dump the code buffer (for debugging)
		code.verify();			
	}

	public static void main (String[] args) {
		junit.textui.TestRunner.run(UEBeispieleTest.class);
	}
}
